/***************************************************************************
 *   Copyright (C) 2005 by Angelo Rosiello                                 *
 *   angelo@rosiello.org                                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 ***************************************************************************/

#include "monitor.h"
#include "syscalls.h"

int singlestep( unsigned long int pid )
{
	/* Keep on tracing syscalls */
        if( ptrace(PTRACE_SINGLESTEP,pid,0L,0L) )
        {
        	ptrace( PTRACE_DETACH,pid,1L,0L );
                perror( "singlestep" );
                return ( -1 );
        }
        waitpid( pid,NULL,WUNTRACED );
}

int monitor( unsigned long int pid, const char* log_filename )
{
	int scall=0, finished=0;
	unsigned long int arg1=0;
	unsigned long int arg2=0;
        unsigned char b4[4];
	unsigned long int value, descriptor;
        char syscall[10];
        char *buffer;
	char *message;
  	register int i;
	FILE *fp;

	//log_filename!=NULL => logging
	if( log_filename!=NULL )
	{
		if( (fp=fopen( log_filename, "aw+" ))==NULL )
		{
			printf( "The Log file cannot be created/opened\n" );
			return( -1 );
		}
	}

	while( !scall )
        {
                if( ptrace(PTRACE_GETREGS,pid,&uregs,&uregs) )
                {
                        ptrace( PTRACE_DETACH,pid,1L,0L );
                        perror( "(rhj) getregs" );
                        return ( -1 );
                }
                *(unsigned long *)&b4[0]=ptrace(PTRACE_PEEKDATA,pid,uregs.eip,&b4[0]);
                if(b4[0]==0xcd && b4[1]==0x80) /* cd 80 = int 80 */
                {
                        scall=1;
                }
                else
                {
			if( singlestep( pid ) == -1 ) return ( -1 );
		}

		if( scall==1 )
                {
                        clean( syscall );
                        sprintf( syscall, "0x%08x", uregs.eax );

			//syscall=SYS_WRITE => monitor it
			if( !strcmp( syscall, sys_write ) )
			{
	                        buffer = (char *)malloc( (uregs.edx+32)*sizeof(char) );
				if( buffer==NULL )
				{
					printf( "(rhj) Fatal Error: Could not allocate memory\n" );
					return( -1 );
				}
        	                clean( buffer );

				//Read the stack
/*
				if( uregs.edx > 2020 )
                		{
                        		printf( "(rhj) Buffer too big\n" );
                        		ptrace( PTRACE_DETACH,pid,1L,0L  );
                        		return ( -1 );
                		}
 */
	              		// Read argument of write
                		for( i=0; i<uregs.edx; i+=4 )
                		{
                        		value=ptrace( PTRACE_PEEKTEXT,pid,uregs.ecx+i,0L );
                        		memcpy( buffer + i, &value, 4 );

                		}
				for( i=0; i<strlen(buffer); i++ ) if( buffer[i]=='\n' )
				{
					buffer[i]='\\';
					buffer[i+1]='n';
					buffer[i+2]='\0';
					buffer[i+3]='\0';
				}

				//Format Output to appear
				message=(char *)malloc( (strlen(buffer)+1024)*sizeof(char) );
				if( message==NULL )
				{
					printf( "(rhj) Fatal Error: Could not allocate memory\n" );
					return( -1 );
				}
				clean( message );
				sprintf( message, "write ( %d, \"%s\", %d );\n", uregs.ebx, buffer, uregs.edx );
				printf( "(rhj) %s", message );
				if( log_filename!=NULL ) log( fp, message );
				free( message );
				free( buffer  );
				if( singlestep( pid ) == -1 ) return ( -1 );
				scall=0;
			}

			//syscall=SYS_READ => monitor it
			else if( !strcmp( syscall, sys_read ) )
			{
	                        buffer = (char *)malloc( (uregs.edx+32)*sizeof(char) );
				if( buffer==NULL )
				{
					printf( "(rhj) Fatal Error: Could not allocate memory\n" );
					return( -1 );
				}
        	                clean( buffer );

				//Read the stack
                		for( i=0; i<uregs.edx; i+=4 )
                		{
                        		value=ptrace( PTRACE_PEEKTEXT,pid,uregs.ecx+i,0L );
                        		memcpy( buffer + i, &value, 4 );

                		}
				for( i=0; i<strlen(buffer); i++ ) if( buffer[i]=='\n' )
				{
					buffer[i]='\\';
					buffer[i+1]='n';
					buffer[i+2]='\0';
					buffer[i+3]='\0';
				}
				//Format Output to appear
				message=(char *)malloc( (strlen(buffer)+1024)*sizeof(char) );
				if( message==NULL )
				{
					printf( "(rhj) Fatal Error: Could not allocate memory\n" );
					return( -1 );
				}
				clean( message );
				sprintf( message, "read  ( %d, \"%s\", %d );\n", uregs.ebx, buffer, uregs.edx );
				printf( "(rhj) %s", message );
				if( log_filename!=NULL ) log( fp, message );
				free( buffer  );
				free( message );
				if( singlestep( pid ) == -1 ) return ( -1 );
				scall=0;
			}

			//syscall=SYS_SOCKET => monitor it
			else if( !strcmp( syscall, sys_socket ) )
			{
	               		//Read argument of SYS_SOCKET
				sprintf( syscall, "0x%08x", uregs.ebx );

				//syscall=SYS_SEND
				if( !strcmp(syscall, sys_send) )
				{
					i=0;
			                while( !finished )
                			{
                        			value=ptrace( PTRACE_PEEKTEXT,pid,uregs.ecx+i,0L );
						if( i==0)	descriptor=value;
						else if( i==4 )
							arg1=value;
						else if( i==8 )
						{
							arg2=value;
						}
						i+=4;
						//If the stack contains a 0x0 the string is finished
						if( value==0L )
						{
							finished=1;
						}

                			}

		                        buffer = (char *)malloc( (arg2+32)*sizeof(char) );
					if( buffer==NULL )
					{
						printf( "(rhj) Fatal Error: Could not allocate memory\n" );
						return( -1 );
					}
					clean( buffer );

                			for( i=0; i<arg2; i+=4 )
                			{
                        			value=ptrace( PTRACE_PEEKTEXT,pid,arg1+i,0L );
                        			memcpy( buffer + i, &value, 4 );
                			}
					message=(char *)malloc( (strlen(buffer)+1024)*sizeof(char) );
					if( message==NULL )
					{
						printf( "(rhj) Fatal Error: Could not allocate memory\n" );
						return( -1 );
					}
					clean( message );

					i=0;
					while( i<strlen(buffer) )
					{
						if( buffer[i]=='\r' )
						{
							buffer[i]='\\';
							buffer[i+2]=buffer[i+1];
							buffer[i+1]='r';
						}

						else if( buffer[i]=='\n' )
						{
							buffer[i]='\\';
						//	buffer[i+2]=buffer[i+1];
							buffer[i+1]='n';
							buffer[i+2]='\0';
							i++;
						}
						i++;
					}

					//Format Output to appear
					//for(i=0; i<arg2; i++) if(buffer[i]=='\n') buffer[i+1]='\0';
					sprintf( message, "send  ( %d, \"%s\", %d );\n", descriptor, buffer, arg2 );
					printf( "(rhj) %s", message );
					if( log_filename!=NULL ) log( fp, message );
					free( buffer  );
					free( message );
					finished=0;
				}

				//sys_call=SYS_RECV
				else if( !strcmp(syscall, sys_recv) )
				{
					i=0;
			                while( !finished )
                			{
                        			value=ptrace( PTRACE_PEEKTEXT,pid,uregs.ecx+i,0L );
						if( i==0)	descriptor=value;
						else if( i==4 )
							arg1=value;
						else if( i==8 )
						{
							arg2=value;
						}

						i+=4;
						//If the stack contains a 0x0 the string is finished
						if( value==0L )
						{
							finished=1;
						}

                			}
		                        buffer=(char *)malloc( (arg2+32)*sizeof(char) );
                		        if( buffer==NULL )
					{
						printf( "(rhj) Fatal Error: Could not allocate memory\n" );
						return( -1 );
					}
					clean( buffer );
                			for( i=0; i<arg2; i+=4 )
                			{
                        			value=ptrace( PTRACE_PEEKTEXT,pid,arg1+i,0L );
                        			memcpy( buffer + i, &value, 4 );
                			}
					//Format Output to appear
					message=(char *)malloc( (strlen(buffer)+1024)*sizeof(char) );
					if( message==NULL )
					{
						printf( "(rhj) Fatal Error: Could not allocate memory\n" );
						return( -1 );
					}
					clean( message );
					i=0;
					while( i<strlen(buffer) )
					{
						if( buffer[i]=='\r' )
						{
							buffer[i]='\\';
							buffer[i+2]=buffer[i+1];
							buffer[i+1]='r';
							i++;
						}

						if( buffer[i]=='\n' )
						{
							buffer[i]='\\';
						//	buffer[i+2]=buffer[i+1];
							buffer[i+1]='n';
							buffer[i+2]='\0';
							i++;
						}
						i++;
					}
					sprintf( message, "recv  ( %d, \"%s\", %d );\n", descriptor, buffer, arg2 );
					printf( "(rhj) %s", message );
					if( log_filename!=NULL ) log( fp, message );
					free( buffer  );
					free( message );
					finished=0;

				}
				if( singlestep( pid ) == -1 ) return ( -1 );
				scall=0;
			}

			//syscall!=(write | read | send | recv) => keep on waiting syscalls
                        else
                        {
                                scall=0;
                                /* Keep on tracing syscalls */
				if( singlestep( pid ) == -1 ) return ( -1 );
			}
		}
	}
	if( fp!=NULL ) fclose( fp );
	return ( 0 );
}

//Logging function
int log( FILE *fp, char *message )
{
	if( fp==NULL )
	{
		perror( "(rhj) Log File" );
		return ( -1 );
	}
	fseek( fp, 0L, SEEK_END );
	fprintf( fp, "%s", message );
	fflush( fp );
	return 0;
}

