/***************************************************************************
 *   Copyright (C) 2005 by Angelo Rosiello                                 *
 *   angelo@rosiello.org                                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 ***************************************************************************/

#include "interactive.h"
#include "parser.h"
#include "help.h"
#include "waiting_syscall.h"
#include "task.h"
#include "monitor.h"
#include "injecting.h"
#include <signal.h>
#include <getopt.h>

const char *program_name;
unsigned long int pid=0;
char *_message;
char *_syscall;
unsigned long int _descriptor=-1;


//OVERVIEW: banner of the program.
//REQUIRES: true
void banner( )
{
	printf( "\nCopyright  Rosiello Security 2005\n" );
	printf( "http://www.rosiello.org\n" );
	printf( "Rosiello Ptrace Injector v1.0\n\n" );
}

void print_usage( FILE* stream, int exit_code )
{
  	banner( );
  	fprintf( stdout, "Usage:  %s options\n", program_name );
  	fprintf( stdout,
           "  -h  --help             Display this usage information.\n"
           "  -l  --log filename     Log to file.\n"
           "  -v  --verbose          Print verbose messages.\n" );
  	exit( exit_code );
}

unsigned long int set_descriptor( char *id )
{
	unsigned long int descriptor;
	descriptor=strtoul( id,NULL,0L );
	printf( "(rhj) Descriptor set to: %d\n", descriptor );
	return ( descriptor );
}

void sigproc(  )
{
	if( pid!= 0 )
	{
		ptrace( PTRACE_DETACH,pid,1L,0L );
		pid=0L;
                _message=NULL;
	        _descriptor = -1;
                _syscall=NULL;
		printf( "Received SIGINT => RESET\n" );
		printf( "Aborting...\n" );
		exit( 0 );
	}
	else
	{
		printf( "Type 'quit' or 'q' to exit.\n" );
	}
}

//OVERVIEW: main of the program. After checking **argv it calls
//	    waiting_syscall( ) and then inject( ).
//REQUIRES: true.
int main( int argc, char **argv )
{
	char *command_line;
	char *command;
	char *argument;
	char *string_help="help";
	char *attach="attach";
	char *detach="detach";
	char *descriptor="descriptor";
	char *message="message";
	char *_monitor="monitor";
	char *_inject="inject";
	char *syscall="syscall";
	int status=0;
	int next_option,i;

  	/* A string listing valid short options letters.  */
  	const char* const short_options = "hl:";
  	/* An array describing valid long options.  */
  	const struct option long_options[] =
	{
    		{ "help",     0, NULL, 'h' },
    		{ "log",      1, NULL, 'l' },
    		{ NULL,       0, NULL, 0   }   /* Required at end of array.  */
  	};

  	/* The name of the file to receive program's log, or NULL for standard output.  */
 	 const char* log_filename = NULL;
  	/* Whether to display verbose messages.  */
  	int verbose = 0;
  	program_name = argv[0];
 	do
	{
    		next_option = getopt_long( argc, argv, short_options, long_options, NULL );
    		switch( next_option )
    		{
    			case 'h':   /* -h or --help */
      			/* User has requested usage information.  Print it to standard
         		output, and exit with exit code zero (normal termination).  */
      			print_usage( stdout, 0 );

    			case 'l':   /* -l or --log */
      			/* This option takes an argument, the name of the log file.  */
      			log_filename = optarg;
      			break;

			case '?':   /* The user specified an invalid option.  */
      			/* Print usage information to standard error, and exit with exit
         		code one (indicating abonormal termination).  */
      			print_usage( stderr, 1 );

    			case -1:    /* Done with options.  */
      			break;

    			default:    /* Something else: unexpected.  */
      			abort ( );
    		}
  	}
  	while( next_option != -1 );

	//Print out the banner
	banner( );

	signal( SIGINT, sigproc );

	/* Core */
	while( 1 )
	{
		command_line = (char *)interactive( pid );
                command = (char *)first_string( command_line );
                argument = (char *)second_string(command_line);

                if( !strcmp(command, string_help) )
                {
                        if( !strcmp(argument, "EMPTY") ) help( );
                        else
		 		command_help( argument );
                }

                else if( !strcmp(argument, "EMPTY") && strcmp(command, string_help)
				&& strcmp(command, _inject) && strcmp(command, _monitor) )
                {
			printf( "(rhj) Bad command: Type help!\n" );
                        command_help( command );
                        command = "NOTHING";
                }

                else if( !strcmp(command, attach) )
                {
                	pid = attaching( argument );
                }

                else if( !strcmp(command, detach) )
                {
			if( pid!=0 )
			{
                        	pid = deattaching( argument );
				printf( "(rhj) Message and descriptor cleaned!\n" );
				_message=NULL;
				_descriptor = -1;
				_syscall=NULL;
 			}
			else
				printf( "(rhj) Attach to some task first!\n" );
                }

		else if( !strcmp(command, descriptor) || !strcmp(command, "d") )
		{
			if( pid!=0 )
			{
				_descriptor = set_descriptor( argument );
			}
			else
				printf( "(rhj) Attach to some task first!\n" );
		}

		else if( !strcmp(command, message) )
		{
			if( pid!=0 ) 
			{
	                        _message = argument;
        	                printf( "(rhj) Message: %s\n", _message );
			}
			else
				printf( "(rhj) Attach to some task first!\n" );
		}

		else if( !strcmp(command, syscall) )
		{
			if( pid!=0 )
			{
				_syscall = argument;
				printf( "(rhj) Syscall: %s\n", _syscall );
			}
	                else
	                        printf( "(rhj) Attach to some task first!\n" );
		}

		//Monitor all defined syscall
		else if( !strcmp(command, _monitor) )
		{
                        if( pid==0 )
                        {
                                printf( "(rhj) Attach to some task first!\n" );
                      	}
			else
			{
				printf( "(rhj) Analyzing all the defined syscalls...\n" );
				monitor( pid, log_filename );
			}
		}		

		else if( !strcmp(command, _inject) )
                {
              		if ( pid==0  )
                        {
 	                       printf( "(rhj) Attach to some task first!\n" );
                        }
                        else if ( _descriptor==-1 )
                        {
        	                printf( "(rhj) Set the descriptor first!\n" );
                        }
                	else if ( _message==NULL )
                        {
                        	printf( "(rhj) Set the message first!\n" );
                        }

			else if ( _syscall==NULL )
			{
				printf( "(rhj) Set the syscall first!\n" );
			}
                        else
                        {
			        printf( "(rhj) Looking for the right chance...\n" );
				waiting:
	                       	if( !waiting_syscall( _syscall, _descriptor, _message, pid ) )
                                {
                                	status=inject( _syscall, _descriptor, _message, pid);
					if( status==0 )
                                	{
                                		printf( "(rhj) Injection Successful!\n" );
						printf( "(rhj) Detached from task, anything cleaned\n" );
						printf( "(rhj)\n" );
						pid=0;
						_message=NULL;
                                		_descriptor = -1;
                                	}
					//Wrong descriptor or not enough space, call back waiting_syscall( )
					else if ( status==-3 )
					{
						goto waiting;		
					}
                                }
                                else
                                {
                                	printf ( "(rhj) Injection Error\n" );
                                        return ( -1 );
                                }
                        }

		}
		else printf( "(rhj) Unknown command: Type help!\n" );
		if( command!=NULL && strcmp(command, "NOTHING") ) 	free( command );
		if( argument!=NULL && strcmp(argument, "EMPTY") && _syscall==NULL ) 	free( argument );
	}

	return ( 0 );
}
